#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <signal.h>
#include "config.h"

/*
Esercizio3
	un programmino di esempio sull'uso dei socket (in modalita` TCP e
	UDP) e le configurazioni client/server in C

client_udp.c
	programma client, non fa altro che connettersi all'host specificato
	nella commandline (porta hardwired) e, se la connessione e` andata
	a buon fine, spedire un comando specificato nella comandline e
	aspettarsi l'output di questo dalla stessa connessione.

Copyright (C) 1998 Marco Gaiarin (gaio@pluto.linux.it)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*/


int	tout = 0;	// c'e` stato un timeout?!


// routine di servizio del SIGALRM
void toutserv(void) {

#ifdef DEBUG
	printf("\t\tDRIIINNN!!!\n");
#endif
	tout = 1;

	// stupido ma questo e` il ``behaviour'' di signal...
	signal(SIGALRM, (void *)toutserv);

}


void main( int argc, char *argv[]) {

char	cmd[CLEN] = "\0",		// comando
	host[MAXHOSTNAMELEN] = "\0",	// host
	buf[BUFLMAX+1],			// buffer per I/O
	seq;				// sequence number di pacchetto

int	sd,				// socket descriptor
	sas,				// socket address size
	bufl = 0,			// quanto letto dal/sul buffer
	a1,a2,a3,a4,			// per eventuali host numerici
	ttime,				// timeout time
	toutc = 0;			// timeouts count value

struct	sockaddr_in sa;			// indirizzo del socket
struct	hostent *he = NULL;		// informazioni sull'host


// un po' di controlli sui parametri passati...
if ( argc != 3) {
	help();
	printf("\nnumero di parametri errato\n");
	exit(OERR);
} else {
	strncat((char *)&host,argv[1],MAXHOSTNAMELEN-1);
	strncat((char *)&cmd,argv[2],CLEN-1);
}
#ifdef DEBUG
printf("client: comando %s, host %s\n",cmd,host);
#endif


// smanetto un po' con i parametri...
if (sscanf(host,"%d.%d.%d.%d",&a1,&a2,&a3,&a4) == 4)
	// sembra proprio un host numerico...
	sa.sin_addr.s_addr=inet_addr(host);
else if ( (he = gethostbyname(host)) == NULL ) {
	perror("client: non riesco a determinare l'indirizzo!");
	exit (OCERR);
}
sa.sin_family 	= AF_INET;
sa.sin_port	= htons(PORT);
memcpy(&sa.sin_addr,he->h_addr,he->h_length);
sas = sizeof(sa);
#ifdef DEBUG
printf("client: UDP, host %s (IP %s), porta %d\n",host,inet_ntoa(sa.sin_addr),(int)ntohs(sa.sin_port));
#endif

// creo un socket
if ( (sd = socket (AF_INET, SOCK_DGRAM, 0)) < 0 ) {
	perror("client: erore facendo socket");
	exit(CDERR);
}
#ifdef DEBUG
printf("client: socket aperto, SD=%u\n",sd);
#endif


// ora aggiungiamo il timeout, un bel while esterno con cui
// fare un po' di prove
while (toutc < MAXTOUT) {

// un altro fork, non posso riaprire la connessione dallo stesso PID
switch ( fork() ) {
case	-1:	// errore in questa fork...

#ifdef DEBUG
	printf("client: fork fallita\n");
#endif
	// lo considero come un timeout
	toutc++;

	// niente controlli astrusi
	break;

case	0:	// figlio, eseguo il comando

	// inizializziamo la routine di gestione dei timeout
	signal(SIGALRM, (void *)toutserv);

	// ora scrivo sul socket la stringa di comando...
	if ( sendto(sd, (char *)&cmd, strlen(cmd), 0, (struct sockaddr *)&sa, sas) != strlen(cmd) ){
		perror("client(child): errore scrivendo sul socket");
		exit(RWERR);
	}
#ifdef DEBUG
	printf("client(child): scritto comando %s\n", cmd);
#endif

	// leggo quello che mi arriva, pacchettizzato cosi` come lo ho spedito
#ifdef DEBUG
	printf("client(child): ricopio (SD=%u)",sd);
#endif
	seq = -1;
	do {
		seq++;
		
		// ``lancio'' un timeout
		ttime = alarm(TOUT);
#ifdef FORCE_TOUT
		if ( (bufl == BUFLMAX+1) && (toutc < MAXTOUT-2) ) {
			printf("forzo timeout\n");
			sleep(TOUT+1);
		}
#endif
#ifdef	FORCE_OOS
		if ( (bufl == BUFLMAX+1) && (toutc < MAXTOUT-1) ) {
			printf("forzo out-of-sequence\n");
			seq++;
		}
#endif
		
		// prendo quello che arriva
		bufl = recvfrom(sd, (char *)&buf, BUFLMAX+1, 0, (struct sockaddr *)&sa, &sas);
#ifdef DEBUG
		printf("%u (n %u, r %u, %u s) ", bufl, seq, buf[0], ttime);
#endif
		// timeout!
		if ( tout == 1 )
			break;

		// out-of-sequence
		if ( buf[0] != seq )
			break;

		// errore!
		if ( bufl < 0 ) {
			perror("client(child): errore leggendo dal socket");
			exit(RWERR);
		}

		write(STDOUT, &buf[1], bufl-1);
	} while ( bufl == BUFLMAX+1 );
#ifdef DEBUG
	printf("dal socket (IP %s)\n", inet_ntoa(sa.sin_addr));
#endif

	if ( tout == 1 )
		exit(1);
	if ( buf[0] != seq )
		exit(2);
	if ( bufl >= 0 )
		exit(0);

	perror("client(child): errore, non dovrei essere qui");
	exit(OERR);

default:	// il padre!!! ;)

	// aspetto di vedere come e` andata...
	wait(&a1);

#ifdef DEBUG
	printf("client: figlio uscito con %u\n",WEXITSTATUS(a1)); 
#endif

	break;

} // switch del fork()

	// devo controllare perche` sono uscito...
	if ( WEXITSTATUS(a1) == 0 )
		break;
	else if ( WEXITSTATUS(a1) == 1 ) {
		printf("\nclient: timeout, riprovo (era %u volta)\n",toutc);
		tout = 0;
		toutc++;
	} else if ( WEXITSTATUS(a1) == 2 ) {
		printf("\nclient: out-of-sequence, riprovo (era %u volta)\n",toutc);
		toutc++;
	} else {
		perror("client: qualche errore");
		exit(OERR);
	}	

// resetto la connessione, altrimenti non la posso riaprire...
shutdown(sd,2);

} // while per timeout

// chiudo il socket
close(sd);

#ifdef DEBUG
printf("client: esco\n");
#endif

} // main()
