#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "config.h"

/*
Esercizio3
	un programmino di esempio sull'uso dei socket (in modalita` TCP e
	UDP) e le configurazioni client/server in C

server-wrapper.c
	semplice wrapper che permette di rimuovere il socket se il server
	reale muore per qualche motivo.
	Tramide #ifdef con questo sorgente compilo ambedue le versioni
	per TCP e UDP

Copyright (C) 1998 Marco Gaiarin (gaio@pluto.linux.it)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*/




void main(void) {

int 	s,				// stato di ritorno
	sas,				// socket address size
	sd;				// socket descriptor
size_t	len = MAXHOSTNAMELEN-1;		// lunghezza hostname
char	hn[MAXHOSTNAMELEN] = "\0";	// hostname in stringa

struct	sockaddr_in sa;			// indirizzo del socket
struct	hostent *he;			// informazioni sull'host


#ifdef TCP
printf("server: TCP, ");
#endif
#ifdef UDP
printf("server: UDP, ");
#endif

// cerco di determinare il mio hostname...
if ( gethostname(hn,len) < 0 ) {
	perror("server-wrapper: non riesco a determinare il mio hostname!");
	exit (OCERR);
}
printf("host %s ",hn);

// cerco di determinare il mio indirizzo...
if ( (he = gethostbyname(hn)) == NULL ) {
	perror("server-wrapper: non riesco a determinare il mio indirizzo!");
	exit (OCERR);
}

// ora smanettiamo per avere la nostra struttura sa ben riempita
sa.sin_family 	= AF_INET;
sa.sin_port	= htons(PORT);
memcpy(&sa.sin_addr,he->h_addr,he->h_length);
sas = sizeof (sa);

printf("(IP %s), porta %d\n",inet_ntoa(sa.sin_addr),(int)ntohs(sa.sin_port));

// creo un socket
#ifdef TCP
if ( (sd = socket (AF_INET, SOCK_STREAM, 0)) < 0 ) {
#endif
#ifdef UDP
if ( (sd = socket (AF_INET, SOCK_DGRAM, 0)) < 0 ) {
#endif
	perror("server-wrapper: erore facendo socket");
	exit(CDERR);
}
#ifdef DEBUG
	printf("server: socket aperto, SD=%u\n",sd);
#endif


// ora lo bindo...
if ( bind(sd, (struct sockaddr *)&sa, sas) < 0 ) {
	perror("server-wrapper: errore facendo bind\n");
	exit(CDERR);
}
#ifdef DEBUG
	printf("server: socket bindato\n");
#endif


// non devo fare molto, solo una fork/wait per il vero server...
switch ( fork() ) {

case 	-1:	// errore!

	perror("server-wrapper: errore facendo fork");
	exit(FWERR);

case	0:	// figlio: eseguo il server vero

	server_net(sd);

default:	// il padre: aspetto la morte del server...

	wait(&s);

} // switch

// non serve fare assolutamente nulla...
printf("server: terminato\n");

shutdown(sd,2);
close(sd);

exit(s);

} // main()
