#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>
#include "config.h"
#include "parser.h"

/*
Esercizio3
	un programmino di esempio sull'uso dei socket (in modalita` TCP e
	UDP) e le configurazioni client/server in C

server_tcp.c
	server reale, cicla in ascolto su un socket e appena sente una
	connessione forka ed elabora la conessione sul figlio.
	il figlio prende in input dal socket un comando e, se il comando
	e` buono, eseguiendo tramite exec il comando stesso; l'output
	del comando e` inviato al client, tramite lo stesso socket.

Copyright (C) 1998 Marco Gaiarin (gaio@pluto.linux.it)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*/


// definisce i parametri di listen e poi aspetta connessioni in maniera
// indefinita, ciclando e producendo figli che andranno a gestire
// la connessione vera e propria, cioe` aspettano input dal socket e se
// quello che arriva e` un comando buono lo eseguono
// i comandi buoni sono quelli definiti in config.h
void server_net(int sd) {

int	csd,			// child socket descriptor
	cmdn,			// numero del comando
	fen = 0,		// Fork Error Number
	i;
char	buf[BUFLMAX],		// buffer per il comando
	cmd[CLEN] = "\0";	// comando da eseguire

struct	sockaddr_in csa;	// indirizzo del socket del figlio
int	csas;			// child socket address size

struct	sigaction infanticida =
	{ SIG_IGN, 0,
	  SA_NOCLDSTOP | SA_RESTART,
	  NULL };		// ammazzo i figli zombies


// ora definisco i parametri di listen...
if ( listen(sd, MAXCONN) < 0 ) {
	perror("server: errore ponendo in modalita` listen");
	exit(OCERR);
}
#ifdef DEBUG
printf("server: modalita` listen attivata\n");
#endif

// un po' di inizalizzazioni...
csa.sin_family = AF_INET;
csa.sin_addr.s_addr = INADDR_ANY;
csa.sin_port = htons(PORT);
csas = sizeof(csa);

// (ri)definisco come handler di SIGCHILD NULL, permettendo ai miei
// poveri figli diventati zombie di riposare in pace
sigaction(SIGCHLD, &infanticida, NULL);

// il server, come tutti i server, non fa altro che fare un bel loop
// aspettando comandi da eseguire
while ( 18 == 18 ) {	// mio giorno di nascita ;)

// ok, mi metto in ascolto
// se passo ottengo un file descriptor da usare normalmente...
if ( (csd = accept(sd, (struct sockaddr *)&csa, &csas)) < 0 ) {
	continue;	// non e` critico, riciclo!!!
} 
#ifdef DEBUG
printf("server: client (da IP %s) accettato, SD=%u\n",inet_ntoa(csa.sin_addr),csd);
#endif


// facciamo un fork al cui interno gestiamo le comunicazioni
switch ( fork() ) {

case 	-1:	// errore!

	// un trucchetto
	if ( fen < MAXFEN ) {
		fen++;
		break;
	} else {
		perror("server: errore facendo fork");
		exit(FWERR);
	}

case	0:	// figlio!

	// (ri)definisco come handler di SIGCHILD NULL, permettendo ai miei
	// poveri figli diventati zombie di riposare in pace
	sigaction(SIGCHLD, &infanticida, NULL);

	// non mi serve...
	shutdown(sd,2);
	close(sd);

	// leggo il comando, azzero il buffer per evitare di trovare
	// schifezze dentro...
	memset(&buf,0,BUFLMAX);
	if ( (i=read(csd, &buf, BUFLMAX-1)) < 0 ) {
		perror("server(child): errore leggendo dal socket");
		exit(RWERR);
	}
	strncpy((char *)&cmd, (char *)&buf, CLEN-1);
	strcat((char *)&cmd, "\0");

	// parso il comando
	cmdn = parse((char *)&cmd);
	if (cmdn == 0)
		printf("server(child): ricevuto comando numero %u, eseguo help()\n",cmdn);
	else
		printf("server(child): ricevuto comando numero %u, eseguo %s\n",cmdn,(char *)getcmd(cmdn));


	// ulteriore fork per il comando...
	switch ( fork() ) {
	case	-1:	// errore in questa fork...
	
		// segnalo l'errore ad ambedue
		perror("server(child): errore facendo fork");
		fprintf((FILE *)&csd,"server(child): errore eseguendo il comando, fork fallita\n");
		
		// niente controlli strani in questa fork
		break;

	case	0:	// figlio, eseguo il comando

		// ok, faccio il cattivo
		close(STDOUT);
		dup(csd);
		close(csd);
		close(STDIN);
		close(STDERR);

		// eseguo il comando vero e proprio
		// devo usare exit, il figlio si deve fermare qui
		exit(docmd(cmdn));

	default:	// padre, devo solo aspettare...

		wait(&i);
#ifdef DEBUG
		printf("server(child): eseguito comando, ha ritornato %u\n",i);
#endif
		break;

	} // switch per exec()

	// chiudo, non mi serve piu`
	shutdown(csd,2);
	close(csd);

	// for sanity, se sono qui tutto e` andato a buon fine
	exit(0);

default:	// il padre

	// un fork e` andato a buon fine, azzero il contatore delle fork
	// andate male
	fen=0;
	
	// for sanity
	break;
	
} // switch del fork()

// lo chiudo, tanto tra un po' lo riapro... ;)))
close(csd);


} // while (18 == 18)

// chiaramente, per ogni loop infinito che si rispetti... ;)))
perror("server: errore loop, non dovrei essere qui");
exit(OERR);

} // server_tcp
