#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <signal.h>
#include <unistd.h>
#include "config.h"
#include "parser.h"

/*
Esercizio3
	un programmino di esempio sull'uso dei socket (in modalita` TCP e
	UDP) e le configurazioni client/server in C

server_udp.c
	server reale, cicla in ascolto su un socket e appena sente una
	connessione forka ed elabora la conessione sul figlio.
	il figlio prende in input dal socket un comando e, se il comando
	e` buono, eseguiendo tramite exec il comando stesso; l'output
	del comando e` inviato al client, tramite lo stesso socket.

Copyright (C) 1998 Marco Gaiarin (gaio@pluto.linux.it)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*/


// routine di servizio del SIGALRM
void toutserv(void) {

#ifdef DEBUG
	printf("\t\tDRIIINNN!!!\n");
#endif

}


// questo e` il server vero e proprio.
// definisce i parametri di listen e poi aspetta connessioni in maniera
// indefinita, ciclando e producendo figli che andranno a gestire
// la connessione vera e propria, cioe` aspettano input dal socket e se
// quello che arriva e` un comando buono lo eseguono
// i comandi buoni sono quelli definiti in config.h
void server_net(int sd) {

int	csd,			// child socket descriptor
	cmdn,			// numero del comando
	fen = 0,		// Fork Error Number
	bufl;			// lunghezza di quanto letto

char	buf[BUFLMAX+1],		// buffer per il comando
				// +1 per il numero di sequenza
	seq,			// sequence number
	cmd[CLEN] = "\0";	// comando da eseguire

struct	sockaddr_in csa;	// indirizzo del socket del figlio
int	upd[2],			// unnamed pipe descriptor
	csas;			// child socket address size

struct	sigaction infanticida =
	{ SIG_IGN, 0,
	  SA_NOCLDSTOP | SA_RESTART,
	  NULL };		// ammazzo i figli zombies


// un po' di inizializzazioni
csa.sin_family = AF_INET;
csa.sin_addr.s_addr = INADDR_ANY;
csa.sin_port = htons(PORT);
csas=sizeof(csa);

// (ri)definisco come handler di SIGCHILD NULL, permettendo ai miei
// poveri figli diventati zombie di riposare in pace
sigaction(SIGCHLD, &infanticida, NULL);


// il server, come tutti i server, non fa altro che fare un bel loop
// aspettando comandi da eseguire
while ( 18 == 18 ) {	// mio giorno di nascita ;)

// con UDP uso recvfrom per ottenere
// oltre ai dati anche l'indirizzo del mittente
memset(&buf,0,BUFLMAX);
if ( recvfrom(sd,(char *)&buf,BUFLMAX,0,(struct sockaddr *)&csa,&csas) < 0 ) {
	// per niente grave, riciclo
	continue;
}
#ifdef DEBUG
printf("server: client (IP %s) ha spedito %s\n",inet_ntoa(csa.sin_addr),buf);
#endif


// facciamo un fork al cui interno gestiamo le comunicazioni
switch ( fork() ) {

case 	-1:	// errore!

	// un trucchetto
	if ( fen < MAXFEN ) {
		fen++;
		break;
	} else {
		exit(FWERR);
		perror("server: errore facendo fork");
	}

case	0:	// figlio!

	// non mi serve...
	close(sd);

	// creo il nuovo socket
	if ( (csd = socket (AF_INET, SOCK_DGRAM, 0)) < 0 ) {
		perror("server(child): erore facendo socket");
		exit(CDERR);
	}
#ifdef DEBUG
	printf("server(child): child socket aperto, SD=%u\n",csd);
#endif

	// trasferisco la parte che mi interessa del buffer
	strncpy((char *)&cmd,(char *)&buf,CLEN-1);
	strcat((char *)&cmd,"\0");

	// parso il comando
	cmdn = parse((char *)&cmd);
	if (cmdn == 0)
		printf("server(child): ricevuto comando numero %u, eseguo help()\n",cmdn);
	else
		printf("server(child): ricevuto comando numero %u, eseguo %s\n",cmdn,(char *)getcmd(cmdn));

	// ok, creo la pipe
	if ( pipe(upd) ) {
		perror("server(child): errore facendo pipe()");
		exit(CDERR);
	}
#ifdef DEBUG
	printf("server(child): unnamed pipe aperta, UPD=%u,%u\n",upd[0],upd[1]);
#endif

	// ulteriore fork per il comando...
	switch ( fork() ) {
	case	-1:	// errore in questa fork...
	
		// segnalo l'errore ad ambedue
		perror("server(child): errore facendo fork");
		fprintf((FILE *)&upd[1],"server(child): errore eseguendo il comando, fork fallita\n");

		// niente controlli astrusi
		break;

	case	0:	// figlio, eseguo il comando

		// ok, faccio il cattivo
		close(STDOUT);
		dup(upd[1]);
		close(upd[1]);
		close(STDIN);

		// eseguo realmente il comando
		exit( docmd(cmdn) );

	default:	// padre, devo spedire sul serio...

#ifdef DEBUG
		printf("server(child): mio figlio sta facendo exec(), io leggo...\n");
#endif

		// E` bene che non mi metta ad aspettare il figlio, altrimenti
		// la pipe si riempie... aspetto un po' e poi inizio a
		// leggere...
		// devo lasciargli il tempo di scrivere almeno un buffer
		// pieno...
		usleep(500*DELAY);
		
		// for sanity
		break;

	} // switch interno

	// inizializziamo la routine di gestione dei timeout
	// mi serve perche` la lettura da unnamed pipe e` tonta, e se per
	// sfiga devo leggere un multiplo intero di MAXBUFL si pianta li
	// come un chiodo...
	// se usato si ridefinisce da solo, ma a me basta funzioni una volta
	signal(SIGALRM, (void *)toutserv);

#ifdef DEBUG
	printf("server(child): ricopio ");
#endif
	seq = 0;
	do {
		// ``lancio'' un timeout per prevenire lo stallo in lettura
		// due secondi sono sufficienti
		alarm(2);
		bufl = read(upd[0], &buf[1], BUFLMAX);
		buf[0] = seq;
		if ( bufl < 0 ) {
			// devo spedire qualcosa, altrimenti di la
			// si arrabbiano...
			buf[1] = '\0';
			bufl = 1;
		}
		// aspetto un po' e poi scrivo... devo usare un delay perche`
		// altrimenti saturo il client se c'e` una grossa quantita`
		// di dati
		// questo usleep() non e` indispensabile ma aiuta
		usleep(DELAY);
		sendto(csd,(char *)&buf,bufl+1,0,(struct sockaddr *)&csa,csas);
		seq++;
#ifdef DEBUG
		printf("%u ",bufl);
#endif
	} while ( bufl == BUFLMAX );
#ifdef DEBUG
	printf("dalla pipe al socket (IP %s)\n",inet_ntoa(csa.sin_addr));
#endif

	close(upd[0]);
	close(upd[1]);
	close(csd);

	// ok, se sono qui il figlio ha fatto il suo mestiere...
	exit(0);

default:	// il padre

	// un fork e` andato a buon fine, azzero il contatore dele fork
	// andate male
	fen=0;	

	// for sanity
	break;
	
} // switch del fork()


} // while (18 == 18)

// chiaramente, per ogni loop infinito che si rispetti... ;)))
perror("server: errore loop, non dovrei essere qui");
exit(OERR);

} // server_udp
